import logging
import typing

from flask import Request

from .formresponse import FormResponse


class Event:
    event_id: str
    event_type: str
    form_response: FormResponse

    def __init__(self, event_id: str, event_type: str,
                 form_response: FormResponse):
        self.event_id = event_id
        self.event_type = event_type
        self.form_response = form_response

    def __str__(self):
        return f'Event {self.event_id} ({self.event_type}): \n {self.form_response}'

    def __repr__(self):
        return self.__str__()

    @classmethod
    def from_request(cls, request: Request, shared_secret: typing.Optional[str] = None):
        if shared_secret is not None:
            # Need to verify this is a valid response using data from the headers
            if not Event._verify_event_with_shared_secret(request=request, shared_secret=shared_secret):
                event_id = request.get_json().get('event_id', None)
                raise ConnectionError(f"Could not verify incoming request with event_id {event_id}")

        return cls.from_event_json(event_json=request.get_json())

    @classmethod
    def from_event_json(cls, event_json: dict):
        event_id = event_json.get('event_id', None)
        event_type = event_json.get('event_type', None)
        form_response_json = event_json.get('form_response', None)
        form_response = FormResponse.from_form_response_json(form_response_json)

        return cls(event_id=event_id, event_type=event_type, form_response=form_response)

    @staticmethod
    def _verify_event_with_shared_secret(request: Request, shared_secret: str) -> bool:
        import hmac
        import base64
        from hashlib import sha256
        payload = request.get_data()
        event_id = request.get_json().get('event_id', None)

        # Verify that the HMAC hash matches the body data
        provided_signature = request.headers.get('typeform-signature')
        digest = hmac.new(shared_secret.encode('utf-8'), payload, sha256).digest()
        calculated_signature = f"sha256={base64.b64encode(digest).decode()}"
        logging.debug(f"Calculated signature {calculated_signature} for event_id {event_id}")

        if calculated_signature == provided_signature:
            logging.info(f'Verified shared secret on incoming request; '
                         f'request with event_id {event_id} is legitimate.')
            return True
        else:
            logging.info(f'Could not verify shared secret on incoming request; '
                         f'request with event_id {event_id} may not be legitimate. '
                         f'Calculated signature: {calculated_signature}, but received signature {provided_signature}')
            return False



import datetime
import logging
import typing

import iso8601

from .answer import Answer
from .field import Field
from .variable import Variable


class FormResponse:
    form_id: str
    token: str
    landed_at: datetime.datetime
    submitted_at: datetime.datetime

    # hidden_variables: dict
    variables: typing.Optional[typing.Dict[str, Variable]] = None
    fields: typing.Dict[str, Field] = None
    answers: typing.Dict[str, Answer]
    _answer_reference_to_id = typing.Dict[str, str]

    def __init__(self, form_id: str, token: str,
                 landed_at: datetime.datetime,
                 submitted_at: datetime.datetime,
                 variables: typing.List[dict],
                 answers: typing.List[dict],
                 fields: typing.List[dict]):
        self.form_id = form_id
        self.token = token
        self.landed_at = landed_at
        self.submitted_at = submitted_at

        # clean up the variables
        self.variables = FormResponse._clean_variable_list(variables) if variables else None

        # clean up the answers
        self.answers, self._answer_reference_to_id = FormResponse._clean_answer_list(answer_list=answers)

        # clean up the field definitions
        self.fields = FormResponse._clean_field_list(field_list=fields)

    @staticmethod
    def _clean_field_list(field_list: typing.List[dict]) -> typing.Dict[str, Field]:
        field_dict: typing.Dict[str, Field] = dict()
        for field in field_list:
            f = Field.from_definition_json(field)
            field_dict[f.id] = f

        return field_dict

    ### Answers
    # Answers are accessible via ID or reference
    @staticmethod
    def _clean_answer_list(answer_list: typing.List[typing.Dict[str, typing.Any]]) -> (typing.Dict[str, Answer],
                                                                                       typing.Dict[str, str]):
        """
        Cleaned answers take the format of
        {
            'field_id_from_typeform': {
                'simple_type': 'choices',
                'full_type': 'multiple_choice',
                'answer': 123, # or 'my_string' or '['string_1', 'string_3']
                'reference': 'block_reference', # if exists, otherwise equals field_id_from_typeform
            },
            ...
        }
        """

        answer_dict = dict()
        reference_to_id = dict()
        for answer in answer_list:
            field: typing.Dict[str, typing.Any] = answer['field']
            field_id = field['id']
            field_type = field['type']
            field_ref = field.get('ref', None)
            if field_ref is None:
                field_ref = field_id

            simple_type = answer['type']

            if simple_type == 'choice':
                if 'label' in answer['choice']:
                    extracted_answer = answer['choice']['label']
                elif 'other' in answer['choice']:
                    extracted_answer = answer['choice']['other']
                else:
                    logging.warning(f'No label or other found for answer {answer}, skipping..')
                    continue
            elif simple_type == 'choices':
                extracted_answer = list()
                if 'labels' in answer['choices']:
                    extracted_answer += answer['choices']['labels']
                if 'other' in answer['choices']:
                    extracted_answer += list(answer['choices']['other'])
            elif simple_type == 'number':
                extracted_answer = answer['number']
            elif simple_type == 'date':
                date_string: str = answer['date']
                extracted_answer = datetime.datetime.strptime(date_string, '%Y-%m-%d')
            else:
                extracted_answer = answer[simple_type]

            # TODO: add Field data like title, properties, etc.
            answer_dict[field_id] = Answer(answer=extracted_answer, full_type=field_type,
                                           reference=field_ref, simple_type=simple_type)
            reference_to_id[field_ref] = field_id

        return answer_dict, reference_to_id

    def get_answer_from_field_id(self, _id) -> Answer:
        return self.answers[_id]

    def get_field_id_from_field_reference(self, reference):
        return self._answer_reference_to_id[reference]

    @property
    def field_references(self) -> typing.List[str]:
        """
        Returns a list of all answer references
        :return: list of answer references
        """
        return list(self._answer_reference_to_id.keys())

    @property
    def field_ids(self) -> typing.List[str]:
        """
        Returns a list of all fields IDs
        :return: list of fields IDs
        """
        return list(self.answers.keys())

    def get_answer_from_field_reference(self, reference, include_field=True) -> Answer:
        """
        Returns the answer from the reference
        :param reference: the reference for this answer
        :return: the Answer object, if found
        """
        _id = self.get_field_id_from_field_reference(reference=reference)
        answer = self.get_answer_from_field_id(_id=_id)
        # TODO: add Field object into the Answer object

        return answer

    @property
    def variable_keys(self) -> typing.List[str]:
        """
        Returns a list of all variable keys
        :return: all variable keys
        """
        return list(self.variables.keys())

    def get_variable_from_key(self, key: str) -> Variable:
        """
        Returns a Variable object from the key
        :param key: the key for this Variable object
        :return: the Variable object
        """
        variable = self.variables[key]
        if variable.key != key:
            raise IndexError(f'Variable key in object ({variable.key} doesnt match dictionary key ({key})). '
                             f'Malformed dictionary?')
        return self.variables[key]

    def get_variable_value(self, key: str) -> typing.Union[str, float, int]:
        """
        Returns the value of the variable from the key
        :param key: the key for the Variable object
        :return: the value stored in this variable
        """
        return self.get_variable_from_key(key).value

    @staticmethod
    def _clean_variable_list(variable_list: typing.Iterable[dict]) -> typing.Dict[str, Variable]:
        clean_dict = dict()
        for variable in variable_list:
            v = FormResponse._clean_variable_single(variable)
            clean_dict[v.key] = v
        return clean_dict

    @staticmethod
    def _clean_variable_single(variable: dict) -> Variable:
        key = variable['key']
        type = variable['type']
        value = variable[type]

        return Variable(key=key, _type=type, value=value)

    def __str__(self):
        return f'FormResponse(form_id={self.form_id}, token={self.token}, ' \
               f'landed_at={self.landed_at}, submitted_at={self.submitted_at},' \
               f'\nvariables={self.variables},\nanswers={self.answers})'

    @classmethod
    def from_form_response_json(cls, form_response_json):
        form_id = form_response_json['form_id']
        token = form_response_json['token']

        landed_at = iso8601.parse_date(form_response_json['landed_at'])
        submitted_at = iso8601.parse_date(form_response_json['submitted_at'])

        variables = form_response_json.get('variables', None)
        answers = form_response_json.get('answers', None)
        fields = form_response_json.get('definition', None).get('fields', None)

        return cls(form_id=form_id, token=token,
                   landed_at=landed_at, submitted_at=submitted_at,
                   variables=variables,
                   answers=answers,
                   fields=fields)



import datetime
import typing

from .field import Field


class Answer:
    answer: typing.Union[typing.List[str], int, float, str]
    full_type: str
    reference: str
    simple_type: str
    field: Field

    def __init__(self, answer: typing.Union[typing.List[str], int, float, str, datetime.datetime], full_type: str,
                 reference: str, simple_type: str):
        self.answer = answer
        self.full_type = full_type
        self.reference = reference
        self.simple_type = simple_type

    def __repr__(self):
        return f'Answer(ref: {self.reference}, full type: {self.full_type}, ' \
               f'simple type: {self.simple_type}, answer: {self.answer})'

    def __str__(self):
        return self.__repr__()



import typing

from pypeform import Answer

_full_type = 'test_full_type'
_reference = 'test_reference'
_simple_type = 'test_simple_type'


def check_answer_values(answer_obj: Answer, answer: typing.Union[typing.List[str], int, float, str],
                        full_type: str, reference: str, simple_type: str) -> bool:
    assert answer_obj.answer == answer
    assert answer_obj.full_type == full_type
    assert answer_obj.reference == reference
    assert answer_obj.simple_type == simple_type

    return True


def test_answer_stores_variables():
    answer = 'test_answer'
    answer_obj = Answer(answer=answer, full_type=_full_type, reference=_reference, simple_type=_simple_type)
    check_answer_values(answer_obj=answer_obj, answer=answer, full_type=_full_type,
                        reference=_reference, simple_type=_simple_type)


def test_answer_stores_int_value():
    answer = 1
    answer_obj = Answer(answer=answer, full_type=_full_type, reference=_reference, simple_type=_simple_type)
    check_answer_values(answer_obj=answer_obj, answer=answer, full_type=_full_type,
                        reference=_reference, simple_type=_simple_type)


def test_answer_stores_float_value():
    answer = 3.14
    answer_obj = Answer(answer=answer, full_type=_full_type, reference=_reference, simple_type=_simple_type)
    check_answer_values(answer_obj=answer_obj, answer=answer, full_type=_full_type,
                        reference=_reference, simple_type=_simple_type)


def test_answer_stores_str_value():
    answer = 'string_value'
    answer_obj = Answer(answer=answer, full_type=_full_type, reference=_reference, simple_type=_simple_type)
    check_answer_values(answer_obj=answer_obj, answer=answer, full_type=_full_type,
                        reference=_reference, simple_type=_simple_type)


def test_answer_stores_list_of_str_value():
    answer = ['string_1', 'string_2', 'skip_a_few', 'string_a_million']
    answer_obj = Answer(answer=answer, full_type=_full_type, reference=_reference, simple_type=_simple_type)
    check_answer_values(answer_obj=answer_obj, answer=answer, full_type=_full_type,
                        reference=_reference, simple_type=_simple_type)



from pypeform import Field


def test_field_from_definition_json():
    import json

    with open('data/body_formatted.json') as f:
        event_data = json.load(f)

    fields = event_data['form_response']['definition']['fields']

    for field in fields:
        field = Field.from_definition_json(field)
        assert isinstance(field, Field)



from pypeform import Event, FormResponse


def test_event_from_json():
    import json
    with open('data/body_formatted.json') as f:
        body_data = json.load(f)

    event = Event.from_event_json(event_json=body_data)
    assert isinstance(event, Event)

    # check basic values
    assert event.event_id == '01GAZ4HA6EK6DP36SG6P3Y4ANM'
    assert event.event_type == 'form_response'

    # check types are correct
    assert isinstance(event.form_response, FormResponse)



import typing


class Choice:
    id: str
    label: str

    def __init__(self, id: str, label: str):
        self.id = id
        self.label = label

    def __repr__(self):
        return f'Choice(label: {self.id}, label: {self.label})'

    def __str__(self):
        return self.__repr__()


class Field:
    id: str
    title: str
    type: str
    ref: str
    allow_multiple_selections: bool
    allow_other_choice: bool
    choices: typing.List[Choice]

    def __init__(self, id: str, title: str, type: str, ref: str, allow_multiple_selections: bool = None,
                 allow_other_choice: bool = None, choices: typing.List[Choice] = None):

        supported_field_types = ["multiple_choice", "short_text", "long_text", "number", "date", "email",
                                 "phone_number", "yes_no", "website", "file_upload", "nps", "opinion_scale",
                                 "ranking", "rating", "dropdown", "legal"]
        if type not in supported_field_types:
            raise NotImplementedError(f"{type} is not an accepted type of field. ")

        self.id = id
        self.title = title
        self.type = type
        self.ref = ref
        self.allow_multiple_selections = allow_multiple_selections
        self.allow_other_choice = allow_other_choice
        self.choices = choices

    def __repr__(self):
        return f'Field(id: {self.id}, title: {self.title}, type: {self.type}, ref: {self.ref}, ' \
               f'allow_multiple_selections: {self.allow_multiple_selections}, allow_other_choice: {self.allow_other_choice}, ' \
               f'choices: {self.choices})'

    @classmethod
    def from_definition_json(cls, field_dict: typing.Dict[str, typing.Any]):
        choice_objs = list()
        if 'choices' in field_dict:
            for choice in field_dict['choices']:
                choice_objs.append(Choice(id=choice['id'], label=choice['label']))

        return cls(id=field_dict['id'],
                   title=field_dict.get('title', None),
                   type=field_dict['type'],
                   ref=field_dict.get('ref', None),
                   allow_multiple_selections=field_dict.get('allow_multiple_selections', None),
                   allow_other_choice=field_dict.get('allow_other_choice', None),
                   choices=choice_objs if len(choice_objs) >= 1 else None
                   )



from pypeform import FormResponse, Variable, Answer, Field


def test_formresponse_from_json():
    import json
    from datetime import datetime
    import pytz

    with open('data/body_formatted.json') as f:
        body_data = json.load(f)

    form_response_json = body_data['form_response']

    fr = FormResponse.from_form_response_json(form_response_json=form_response_json)

    assert isinstance(fr, FormResponse)

    # check basic values
    assert fr.form_id == 'sRmK2H1s'
    assert fr.token == '7ax1kv4zahmlj1d9m7ax150zqknp9jci'
    assert fr.landed_at == datetime(2022, 8, 21, 2, 41, 32, tzinfo=pytz.UTC)
    assert fr.submitted_at == datetime(2022, 8, 21, 2, 43, 28, tzinfo=pytz.UTC)

    # check Variable types
    assert isinstance(fr.variables, dict) or fr.variables is None
    if fr.variables is not None:
        assert all(isinstance(key, str) for key in fr.variables.keys())
        assert all(isinstance(value, Variable) for value in fr.variables.values())

    # check Answer types
    assert isinstance(fr.answers, dict)
    assert all(isinstance(key, str) for key in fr.answers.keys())
    assert all(isinstance(value, Answer) for value in fr.answers.values())

    # check Field types
    assert isinstance(fr.fields, dict)
    assert all(isinstance(key, str) for key in fr.fields.keys())
    assert all(isinstance(value, Field) for value in fr.fields.values())



import typing


class Variable:
    key: str
    type: str
    value: typing.Union[str, float, int]

    def __init__(self, key: str, _type: str, value: typing.Union[str, float, int]):
        self.key = key
        self.type = _type
        self.value = value

    def __repr__(self):
        return f'Variable(key: {self.key}, type: {self.type}, value: {self.value})'

    def __str__(self):
        return self.__repr__()



from pypeform import Variable

test_key = 'test_key'
test_type = 'test_type'


def check_variable_values(variable: Variable, key, type, value) -> bool:
    assert variable.key == key
    assert variable.type == test_type
    assert variable.value == value

    return True


def test_variable_stores_values():
    value = 'test_value'
    variable = Variable(key=test_key, _type=test_type, value=value)
    check_variable_values(variable=variable, key=test_key, type=test_type, value=value)


def test_variable_stores_int_value():
    value = 1
    variable = Variable(key=test_key, _type=test_type, value=value)
    check_variable_values(variable=variable, key=test_key, type=test_type, value=value)


def test_variable_stores_float_value():
    value = 3.14
    variable = Variable(key=test_key, _type=test_type, value=value)
    check_variable_values(variable=variable, key=test_key, type=test_type, value=value)


def test_variable_stores_str_value():
    value = 'test_string'
    variable = Variable(key=test_key, _type=test_type, value=value)
    check_variable_values(variable=variable, key=test_key, type=test_type, value=value)
